"""Custom exceptions for pdpipe."""


class FailedPreconditionError(Exception):
    """An exception raised when a pipeline stage is applied to a dataframe for
    which the stage precondition does not hold.
    """


class FailedPostconditionError(Exception):
    """An exception raised when an expected post-condition does not hold after
    an application of the a pipeline stage to a dataframe.
    """


class FailedConditionError(Exception):
    """An exception raised when an expected condition does not hold for an
    input dataframe.
    """


class PipelineInitializationError(Exception):
    """An exception raised when a pipeline is not initialized properly."""


class PipelineApplicationError(Exception):
    """An exception raised when an exception is thrown during the application
    of a pipeline or a pipeline stage to a dataframe.
    """


class UnfittedPipelineStageError(Exception):
    """An exception raised when a (non-fit) transform is attempted with an
    unfitted pipeline stage.
    """


class UnexpectedPipelineMethodCallError(Exception):
    """An exception raised when a placeholder method implementation of an
    extension of the PdPipeline class is called when it was not expected.

    If this exception is raised, it is very likely a result of a bug in the
    pdipe package. Please open an issue on https://github.com/pdpipe/pdpipe

    An example is a custom pipeline stage class that is an X-y transformer:
    meaning, it takes X and y as input and returns X and y. In order to declare
    itself as such, it implements the `_transform_Xy()` and the
    `_fit_transform_Xy` methods. However, since `_transform` is an abstract
    method of the PdPipeline class, it must be implemented, and so it is
    implemented to raise this error, expecting to never be called.

    Thus, if `_transform` is called instead of `_transform_Xy` for this stage,
    it is unexpected, and this exception is raised.
    """



"""Handles for all pandas.DataFrame dataframe-outputing transformations."""

from typing import List, Dict

from pandas import DataFrame

from ..core import PdPipelineStage


# === DataFrame methods  ===

class _DataFrameMethodTransformer(PdPipelineStage):

    def __init__(
        self,
        method_name: str,
        args: List[object],
        kwargs: Dict[str, object],
    ) -> None:
        self._method_name = method_name
        self._args = args
        self._kwargs = kwargs
        # we must always pop 'inplace', if it's there
        found = self._kwargs.pop('inplace', None)
        if found is not None:
            self._kwargs['inplace'] = False
        exmsg = (
            "Pipeline stage failed while applying method {} with kwargs {}"
        ).format(method_name, self._kwargs)
        desc = "Apply dataframe method {} with kwargs {}".format(
            method_name, self._kwargs)
        super_kwargs = {
            'exmsg': exmsg,
            'desc': desc,
            'name': method_name,
        }
        super().__init__(**super_kwargs)

    def _prec(self, df: DataFrame) -> bool:  # pragma: no cover
        return True

    def _transform(self, df: DataFrame, verbose: bool) -> DataFrame:
        method = getattr(df, self._method_name)
        return method(*self._args, **self._kwargs)


class _DfMethodTransformerHandle():

    def __init__(self, method_name: str, doc: str) -> None:
        self._method_name = method_name
        self.__doc__ = doc

    def __call__(self, *args, **kwargs: Dict[str, object]) -> PdPipelineStage:
        return _DataFrameMethodTransformer(
            method_name=self._method_name,
            args=args,
            kwargs=kwargs,
        )


__RETURNS = 'Returns'
__DATAFRAME = 'DataFrame'


def _is_dataframe_transform(attr_name: str, attr: object) -> bool:
    if attr_name.startswith('_') or not callable(attr):
        return False
    try:
        doc_lines = attr.__doc__.split('\n')
        returns_line_index = None
        for i, line in enumerate(doc_lines):
            if __RETURNS in line:
                returns_line_index = i + 2
                break
        if returns_line_index:
            return_type_line = doc_lines[returns_line_index]
            if __DATAFRAME in return_type_line:
                return True
        return False
    except (AttributeError, IndexError):  # pragma: no cover
        return False


# === END-OF DataFrame methods  ===



"""Whitelist and blacklist of pandas.Series transforms."""


SERIES_TRANSFORMS_WHITELIST = [
    'add', 'add_prefix', 'add_suffix', 'append', 'argsort', 'asof', 'backfill', 'combine_first', 'convert_dtypes', 'copy', 'describe', 'diff', 'div', 'divide', 'drop', 'drop_duplicates', 'dropna', 'duplicated', 'eq', 'explode', 'fillna', 'floordiv', 'ge', 'gt', 'isin', 'isna', 'isnull', 'le', 'lt', 'map', 'mod', 'mode', 'mul', 'multiply', 'ne', 'nlargest', 'notna', 'notnull', 'nsmallest', 'pad', 'pct_change', 'pow', 'radd', 'rdiv', 'reindex', 'reindex_like', 'rename', 'rename_axis', 'replace', 'rfloordiv', 'rmod', 'rmul', 'round', 'rpow', 'rsub', 'rtruediv', 'sample', 'set_axis', 'set_flags', 'shift', 'sort_values', 'squeeze', 'sub', 'subtract', 'truediv', 'value_counts', 'view',  # noqa: E501
    'reset_index', 'combine', 'at_time', 'compare',
    'apply', 'between', 'between_time', 'swaplevel', 'droplevel',
    'dot', 'transform', 'tshift', 'tz_localize',
    'to_period', 'to_timestamp',
    'first', 'kurt', 'kurtosis', 'last', 'mad', 'max', 'mean', 'median', 'min', 'prod', 'product', 'quantile', 'sem', 'skew', 'std', 'sum', 'var', 'xs',  # noqa: E501
]

SERIES_TRANSFORMS_BLACKLIST = [
    'agg', 'aggregate', 'groupby', 'align',
    'count', 'divmod', 'rdivmod',
]
